// ----------------------------------------------------------------------
// FIAT (IAW 16F) Test, (c) 2008, Brainsucker (brainsucker.na@gmail.com).
// ----------------------------------------------------------------------

#include "stdafx.h"

#define DIAGCNT 0x3A

HANDLE m_PortHandle;
OVERLAPPED m_OvrRead;
OVERLAPPED m_OvrWrite;	
__int64 perfFreq;

bool waitIso = true;
TCHAR comDevice[128];

void SetBaudRate(int Baudrate)
{
	PurgeComm(m_PortHandle, PURGE_TXABORT | PURGE_RXABORT | PURGE_TXCLEAR | PURGE_RXCLEAR);
	COMSTAT cs;
	DWORD errs;
    DCB dcb;
    ZeroMemory(&dcb, sizeof(DCB));
    dcb.DCBlength = sizeof(DCB);
    dcb.BaudRate = Baudrate;
    dcb.fBinary = 1;
    dcb.fParity = 0;
    dcb.ByteSize = 8;
    dcb.StopBits = ONESTOPBIT;
    SetCommState(m_PortHandle, &dcb);
	ClearCommError(m_PortHandle, &errs, &cs);
	printf("Port %S BaudRate %d COM Errors %X, comstate %X - in %X - out %X\n", comDevice, Baudrate, errs, *((DWORD*)(&cs)), cs.cbInQue, cs.cbOutQue);
	EscapeCommFunction(m_PortHandle, SETBREAK);
	EscapeCommFunction(m_PortHandle, CLRBREAK);
	EscapeCommFunction(m_PortHandle, SETDTR);
	EscapeCommFunction(m_PortHandle, CLRRTS);
	COMMTIMEOUTS ct;
	ct.ReadIntervalTimeout = ct.ReadTotalTimeoutMultiplier = MAXDWORD; ct.ReadTotalTimeoutConstant = 10;
	ct.WriteTotalTimeoutConstant = ct.WriteTotalTimeoutMultiplier = 0;
	SetCommTimeouts(m_PortHandle, &ct);
}

bool PassedSince(__int64 ms, __int64 from)
{
	__int64 perf1;
	QueryPerformanceCounter((LARGE_INTEGER*)&perf1);
	return (perf1-from) > ((ms * perfFreq)/1000);
}

void WaitFor(__int64 ms, bool silent = false)
{
	if (!silent)
		printf("waiting for %d ms...\n", ms);
	__int64 perf1;
	QueryPerformanceCounter((LARGE_INTEGER*)&perf1);
	while (!PassedSince(ms, perf1)) Sleep(0);
}

void SendByte(char *buf, bool silent = false)
{
	DWORD read;
	WriteFile(m_PortHandle, buf, 1, &read, NULL);
	if (!silent)
		printf("wrote > %02X\n", (DWORD)((unsigned char) *buf));
}

void Send(int v, bool silent = false)
{
	char b = v;
	SendByte(&b, silent);
}

__int64 GetResponse(int b, bool silent = false)
{
	__int64 res = -1;
	DWORD read;
	char cb = (char)b, ocb = 0;
	SendByte(&cb, silent);
	__int64 perf1;
	QueryPerformanceCounter((LARGE_INTEGER*)&perf1);
	while (res == -1 && !PassedSince(300, perf1))
	{
		ReadFile(m_PortHandle, &ocb, 1, &read, NULL);
		if (read > 0)
		{
			res = (unsigned char) ocb;
			if (!silent)
				printf("read > %02X\n", (DWORD)((unsigned char) ocb));
		}
	}
	WaitFor(5, silent);
	return res;
}

void ClosePort()
{
	if (m_PortHandle != NULL)
		CloseHandle(m_PortHandle);
	m_PortHandle = NULL;
}

bool OpenPort()
{
	ClosePort();
	while(true)
	{
		m_PortHandle = CreateFile(comDevice, GENERIC_READ|GENERIC_WRITE, 0, NULL, OPEN_EXISTING, 0, 0);
		if (m_PortHandle != NULL && m_PortHandle != INVALID_HANDLE_VALUE) break;
		printf("couldn't open port (q - quit)...\n");
		Sleep(1000);

		while(_kbhit())
		{
				char c = _getch();
				if (c == 'q' || c == 'Q' || c == 'x' || c == 'X') return false;
		}
	} 
	return true;
}

bool Reset()
{
	printf("\n\nResetting...\n");
	if (!OpenPort()) return false;
	SetBaudRate(1200);
	printf("Waiting for ISO code, R - reset, N - test cable (connect ecu-RX-TX), Q - quit\n");

	unsigned char isobuf[6];
	memset(isobuf, 0, sizeof(isobuf));
	while (waitIso)
	{
		for (int i = 0; i < 5; i++)
			isobuf[i] = isobuf[i+1];
		while (true)
		{
			while(_kbhit())
			{
				char c = _getch();
				if (c == 'r' || c == 'R')
					return Reset();
				if (c == 'q' || c == 'Q' || c == 'x' || c == 'X') return false;
				if (c == 'n' || c == 'N')
				{
					int i = 0;
					while (true)
					{
						if (_kbhit()) break;
						DWORD read, start = GetTickCount();
						char buf[1];
						buf[0] = (char)i++;
						WriteFile(m_PortHandle, buf, 1, &read, NULL);
						printf("%02X <==> ", (DWORD)((unsigned char) buf[0]));
						while (GetTickCount() - start < 200)
						{
							ReadFile(m_PortHandle, buf, 1, &read, NULL);
							if (read > 0)
								printf("%02X", (DWORD)((unsigned char) buf[0]));
							Sleep(10);
						}
						printf("\n");
					}	
					return Reset();
				}

			}
			DWORD read;
			ReadFile(m_PortHandle, &isobuf[5], 1, &read, NULL);
			if (read > 0) break;
			Sleep(1);
		}
		printf("read > %02X\n", (DWORD)((unsigned char) isobuf[5]));
		if (isobuf[1] == 0xD0 && isobuf[2] == 0x85 && isobuf[0] == 0x55) break;
	}
	printf("got iso code: ");
	for (int i = 0; i < 6; i++) printf("%02X ", isobuf[i]);
	printf("\nwaiting for 500 ms to response...\n");
	QueryPerformanceFrequency((LARGE_INTEGER*)&perfFreq);
	printf("performance freq %d\n\n", perfFreq);

	// answer after 500 ms
	WaitFor(550);

	printf("sending handshake response...\n");

	unsigned char answerBytes[3] = {0x0F, 0xAA, 0xCC};
	for (int i = 0; i < sizeof(answerBytes); i++)
	{
		SendByte((char*)&answerBytes[i]);
		WaitFor(100);
	}

	WaitFor(100);
	printf("switching to 7680 rate and diagnostic mode...\n");
	SetBaudRate(7680);

	printf("\n\n");
	return true;
}

void GetDiagInfo(__int64 *diagvalues, bool full, bool silent = false, bool supershort = false)
{
	if (!silent)
		printf("data => ");
	// request diag bytes
	for (int b = 1; b < (full?DIAGCNT:(supershort?5:16)); b++)
	{
			if (b == 0x31) continue;
			__int64 res = GetResponse(b, true);
			if (res == -1)
			{
				diagvalues[b] = 0;
				if (!silent)
					printf("NA ");
			} else
			{
				diagvalues[b] = res;
				if (!silent)
					printf("%02X ", res);
			}			
	}
	if (!silent)
		printf("\n\n");
}

void PrintBasicInfo(__int64 *diagvalues)
{
	printf("RPM: %12I64d   Inj:  %5.3f   IgnA: %10.1f   MAP: %5I64d mmHG\nAir: %12I64d   Wat: %6I64d   Thr: %3I64d (%5.2f)   Bat: %5.2f\nLmd: %3I64d (%6.3f)   ISPos: %4I64d   ISInt: %9I64d   ISPrp: %3I64d   TT: %c\n",
			(diagvalues[1]*256 + diagvalues[2] != 0)?(15000000/(diagvalues[1]*256 + diagvalues[2])):(0),
			0.002*(diagvalues[3]*256 + diagvalues[4]),
			diagvalues[5]/2.0,
			diagvalues[6]*3,
			diagvalues[7]-40,
			diagvalues[8]-40,
			diagvalues[9], diagvalues[9]*0.4234 -2.9638,
			diagvalues[10]*0.0625,
			diagvalues[11], diagvalues[11]*0.002656 + 0.66,
			diagvalues[12], 
			diagvalues[13],
			diagvalues[14],
			(diagvalues[15] == 0)?'-':((diagvalues[15] == 0x80)?('0'):('+')));
}

char *statusInfo[16] = {"Diag OK, stepper OK, TH20 OK", "Motor rotating", "Square Wave Signal OK", "ICV at Min or Max", "Operating in Closed Loop", "Air conditioner ON", "System calibration OK", "Testing Stepper Motor", NULL, NULL, "Stepper Calibration OK", NULL, NULL, NULL, "Power line failure", NULL};
char *errorInfo[24] = {"Throttle Position Sensor", "MAP Sensor", "Lambda Probe", "Water Temperature Sensor", "Air Temperature Sensor", "Battery Voltage", "Estimate Error Steps", NULL,
			"Injector", "Ignition Coil 1", "Ignition Coil 2", "Idle Stepper", "Charcoal Canister", "Air Conditioner", "Fuel Pump", "Generic Relay",
			"Calibration", "RAM", "EPROM", "EEPROM", "Microcontroller", "RPM/TDC Sensor", NULL, NULL};
char *errorOn[24] = {"Shorted to GND", "Shorted to VBatt", NULL, "Shorted to GND", "Shorted to GND", "Larger than upper", NULL, NULL, 
			"Short circuited", "Shorted to VBatt", "Shorted to VBatt", "Short Circuted", "Shorted to VBatt", "Shorted to VBatt", "Shorted to VBatt", "Shorted to VBatt",
			"Parameters Too Rich", NULL, NULL, NULL, NULL, NULL, NULL, NULL};
char *errorOff[24] = {"Shorted to VBatt", "Shorted to GND", NULL, "Shorted to VBatt", "Shorted to VBatt", "Lower than lower", NULL, NULL, 
			"Open circuited", "Shorted to GND", "Shorted to GND", "Open Circuted", "Shorted to GND", "Shorted to GND", "Shorted to GND", "Shorted to GND",
			"Parameters Too Lean", NULL, NULL, NULL, NULL, NULL, NULL, NULL};

void PrintErrorsInfo(__int64 *diagvalues)
{
	__int64 stat, tmp = stat = diagvalues[0x2A]*256 + diagvalues[0x13];
	int c = 0;
	printf("\nFGSTAT:    %I64X      FGSTAT2:    %I64X\nPOSITIVE status bits:\n", diagvalues[0x13], diagvalues[0x2A]);
	for (int i = 0; i < 16; i++, tmp >>= 1)
		if (((tmp & 1) != 0) && statusInfo[i] != NULL) 
		{
			printf("    [%s]", statusInfo[i]);
			if (++c % 3 == 0) printf("\n");
		}
	tmp = stat;
	printf("\nNEGATIVE status bits:\n");
	for (int i = 0; i < 16; i++, tmp >>= 1)
		if (((tmp & 1) != 1) && statusInfo[i] != NULL) 
		{
			printf("  [%s]", statusInfo[i]);
			if (++c % 3 == 0) printf("\n");
		}
	printf("\n\nERRORS:  CO_INP %2I64X VA_INP %2I64X CO_OUT %2I64X VA_OUT %2I64X CO_FUN %2I64X VA_FUN %2I64X\nDETAILS: CO_INP %2I64X VA_INP %2I64X CO_OUT %2I64X VA_OUT %2I64X CO_FUN %2I64X VA_FUN %2I64X\nDECODED ERRORS:\n",
			diagvalues[0x10], diagvalues[0x14], diagvalues[0x11], diagvalues[0x15], diagvalues[0x12], diagvalues[0x16],
			diagvalues[0x2B], diagvalues[0x2E], diagvalues[0x2C], diagvalues[0x2F], diagvalues[0x2D], diagvalues[0x30]);

	__int64 valErrs = diagvalues[0x10] + (diagvalues[0x11] << 8) + (diagvalues[0x12] << 16) + ((diagvalues[0x14] + (diagvalues[0x15] << 8)+ (diagvalues[0x16]<<16)) << 32);
	__int64 valDets = diagvalues[0x2B] + (diagvalues[0x2C] << 8) + (diagvalues[0x2D] << 16) + ((diagvalues[0x2E] + (diagvalues[0x2F] << 8)+ (diagvalues[0x30]<<16)) << 32);

	for (int i = 0; i < 24; i++, valErrs >>= 1, valDets >>= 1)
	{
		__int64 vE = valErrs & (1 + (((__int64)1)<<32)), vD = valDets & (1 + (((__int64)1)<<32));
		if (vE == 0 || errorInfo[i] == NULL) continue;
		bool validated = false;
		while (vE != 0)
		{
			if ((vE & 1) != 0)
			{
				printf("  [%10s]    %s ", validated?"VALIDATED":"FILTERED", errorInfo[i]);
				if (errorOn[i] != NULL || errorOff[i] != NULL)
					printf("(%s)", ((vD & 1) != 0)?errorOn[i]:errorOff[i]);
				printf("\n");
			}
			vE >>= 32; vD >>= 32; validated = !validated;
		}		
	}
	
	char err[2+0x21-0x17];
	for (int i = 0x17; i <= 0x21; i++)
		err[i-0x17] = (char) diagvalues[i];
	err[1+0x21-0x17] = 0;
	printf("\nREPORTED ERROR: %s\n",err);
}

void PrintAdvancedInfo(__int64 *diagvalues)
{
	char err[2+0x21-0x17];
	for (int i = 0x17; i <= 0x21; i++)
		err[i-0x17] = (char) diagvalues[i];
	err[1+0x21-0x17] = 0;
	printf("ISCorr: %3I64d   ISCorrAC: %3I64d   MinIS: %3I64d   ExtCorICV: %3I64d\nSOBM_CCOff: %5I64d   SOBM_CCOn: %5I64d   SOred: %5I64d   SOredMP: %5I64d\nTargRPM: %4I64d   DiffRPM: %4I64d   ERROR: %s\n",
		diagvalues[0x22]*256 + diagvalues[0x23],
		diagvalues[0x24]*256 + diagvalues[0x25],
		diagvalues[0x28]-128,
		diagvalues[0x29]-128,
		diagvalues[0x32]*256 + diagvalues[0x33],
		diagvalues[0x34]*256 + diagvalues[0x35],
		diagvalues[0x36]*256 + diagvalues[0x37],
		diagvalues[0x38]*256 + diagvalues[0x39],
		diagvalues[0x26]*8,
		diagvalues[0x27]*8,
		err);
}

int _tmain(int argc, _TCHAR* argv[])
{
	printf("\nFIAT (IAW 16F) Test, (c) 2008, Brainsucker (brainsucker.na@gmail.com).\nThanks and regards are going to:\n  Chuck Storry (forzavx.com), Will Hoult, OpenDiag yahoo group, FiatForum.com.\n\n");
	if (argc < 2)
	{
		printf("Run as:\n\tFiatTest.exe COM2\nsubstituting COM2 with your right serial port name.\n\nWARNING: This program is only tested to work with Will Hoult scheme\n(see forzavx.com IAW/GECU section).\n\n");
		return 0;		
	}
	lstrcpy(comDevice, argv[1]);
    // Prepare OVERLAPPED structures
    /*memset(&m_OvrRead, 0, sizeof(OVERLAPPED));
    memset(&m_OvrWrite, 0, sizeof(OVERLAPPED));
    m_OvrRead.hEvent = CreateEvent(NULL, TRUE, FALSE, L"");
    m_OvrWrite.hEvent = CreateEvent(NULL, TRUE, FALSE, L"");*/

	if (!Reset()) return 0;

	bool full = false, run = true;
	int i = 0;
	__int64 diagvalues[DIAGCNT];
	memset(diagvalues, 0, sizeof(diagvalues));
	while (run)
	{
		// react to keys
		while (_kbhit())
		{
			char c = _getch();
			switch (c)
			{
			case 'q': case 'Q': case 'x': case 'X': run = false; break;
			case 'f': case 'F': full = !full; break;
			case 'p': case 'P': _getch(); break;
			case 'r': case 'R': if (!Reset()) return 0; break;
			case 's': case 'S':
				{
					printf("Super-short Super-fast RPM/Injection mode...\n");
					while (run)
					{
						if (_kbhit()) break;
						GetDiagInfo(diagvalues, false, true, true);
						__int64 rpm = (diagvalues[1]*256 + diagvalues[2] != 0)?(15000000/(diagvalues[1]*256 + diagvalues[2])):(0);
						double inj = 0.002*(diagvalues[3]*256 + diagvalues[4]);
						double fueluse = ((rpm*2.0)/60.0)*inj;
						printf("%02I64X %02I64X %02I64X %02I64X => RPM:  %4I64d   Injection:  %6.3f   Fuel-use (ms-per-s):  %7.3f\n", 
							diagvalues[0x01],diagvalues[0x02],diagvalues[0x03],diagvalues[0x04], 
							rpm,
							inj, fueluse);
					}
					_getch();
				} break;
			case 'v': case 'V': 
				{
					while (true)
					{
						int val;
						// single value retreival mode
						printf("enter mem index (hex, 0 - leave): ");
						scanf("%X", &val);
						printf("read value 0x%02X (%d)\n", val, val);
						if (val == 0) break;
						__int64 res = GetResponse(val);
						printf("\nmem[%02X] == ", val);
						if (res == -1) printf("NA");
						else printf("%I64X (%I64d)", res, res);
						printf("\n\n");
					}					
				} break;

			case 'e': case 'E':
				{
					GetDiagInfo(diagvalues, true);
					PrintErrorsInfo(diagvalues);
					printf("\npress C to clear errors on next poweron, any key to continue...\n");
					char c = _getch();
					if (c == 'c' || c == 'C')
					{
						bool ok = false;
						printf("Trying to clear errors from ECU memory (MIL OFF)...\n");
						do 
						{
							printf("Entering diag mode...\n");
							__int64 res = GetResponse(0xAA);
							if (res != 0xAA) break;
							printf("Sending clear command...\n");
							res = GetResponse(0x84);
							if (res != 0x84) break;
							printf("Waiting for clear result...\n");
							DWORD read, start = GetTickCount();
							char buf[1];
							while (GetTickCount() - start < 1000)
							{
								ReadFile(m_PortHandle, buf, 1, &read, NULL);
								if (read > 0)
								{
									printf("read> %02X\n", (DWORD)((unsigned char) buf[0]));
									if ((DWORD)((unsigned char) buf[0]) == 0xFF) ok = true;
									break;
								}
								Sleep(10);
							}
							if (!ok) printf("Clear timeout or error...\n");
							printf("Leaving diag mode...\n");
							res = GetResponse(0xFF);
						} while (false);
						if (ok) printf("Errors from ECU memory cleared sucessfully!\n");
						else printf("Failed to clear ECU errors!\n");
						printf("press any key to continue...\n");
						_getch();
					}
				}
				break;
			
			case 't': case 'T': 
				{
					printf("\nEntering test mode (0xAA)...\n");
					__int64 res = GetResponse(0xAA);
					if (res != 0xAA)
					{
						printf("Test mode failed...\n\n");
						break;
					}
					printf("Test mode!\n\n");

					while (true)
					{
						int val;
						// single value retreival mode
						printf("Tests:\n  80 - Fuel Pump     81 - Injector     82/83 - Ignition coil 1 / 2\n"
							"84 - Error cancelation     85 - Charcoal canister     86 - Engine Rev Limiter\n"
							"87 - Air Condition Relay    88 - Generic Relay     89 - Limp Home (AC off/on)\n"
							"8E - Stepper operation    Disabled: 8F - TRIMRAM 91/92 - Stepper AC/Regulation\n");
						printf("enter test index (hex, 0 - leave, valid 80-89, 8E, 91): ");
						scanf("%X", &val);
						printf("test 0x%02X (%d)\n\n", val, val);
						if (val < 0x80 || val > 0xA0) break;

						// initiate test
						int test = (int)GetResponse(val);
						if (test != val)
						{
							printf("failed to start test %02X, response %02X...\n", val, test);
							continue;
						}
						printf("test %02X started, press any key to cancel or wait for ECU result...\n", val);
						int res = 0, t = 0;

						while (res == 0)
						{
							if (_kbhit())
							{
								_getch();
								printf("\n\nInterupting...\n");
								res = (int) GetResponse(0xFF);
								break;
							}
							DWORD read;
							char ocb = 0;
							ReadFile(m_PortHandle, &ocb, 1, &read, NULL);
							if (read > 0)
							{
								res = (unsigned char) ocb;
								printf("\n\nread > %02X\n", (DWORD)((unsigned char) ocb));
							} else 
							{
								printf("%d.%d  ", (++t)/10, t%10);
								Sleep(100);
							}
						}

						printf("\n\nTest result %02X == %s\n\n", res, (res == 0xFF)?("OK"):("FAIL"));
					}					

					printf("\nLeaving test mode (0xFF)...\n");
					res = GetResponse(0xFF);
					if (res != 0xFF)
					{
						printf("Test mode leave failed...\n\n");
						break;
					}
					printf("Test mode LEFT!\n\n");
				} break;
			}
		}
		if (!run) break;

		GetDiagInfo(diagvalues, full, full);

		PrintBasicInfo(diagvalues);
		if (full) PrintAdvancedInfo(diagvalues);
		
		printf("\n");

		/*DWORD read, start = GetTickCount();
		char buf[1];
		buf[0] = (char)i++;
		WriteFile(m_PortHandle, buf, 1, &read, NULL);
		printf("%02X -> ", (DWORD)((unsigned char) buf[0]));
		while (GetTickCount() - start < 200)
		{
			ReadFile(m_PortHandle, buf, 1, &read, NULL);
			if (read > 0)
				printf("%02X ", (DWORD)((unsigned char) buf[0]));
			Sleep(10);
		}
		printf("\n");*/
	}
	ClosePort();
	return 0;
}

